@@ -106,15 +106,17 @@ class SimpleParser(Parser):
 #
 # Default Parser implementation
 #
-_STATE_GROUND = 0
-_STATE_ESC = 1
+_STATE_GROUND           = 0
+_STATE_ESC              = 1
 _STATE_ESC_INTERMEDIATE = 2
-_STATE_CSI_PARAMETER = 3
+_STATE_CSI_PARAMETER    = 3
 _STATE_CSI_INTERMEDIATE = 4
-_STATE_SS2 = 6
-_STATE_SS3 = 7
-_STATE_OSC = 8
-_STATE_STR = 9 
+_STATE_SS2              = 6
+_STATE_SS3              = 7
+_STATE_OSC              = 8
+_STATE_OSC_ESC          = 9
+_STATE_STR              = 10 
+_STATE_STR_ESC          = 11
 
 class _MockHandler:
 
@@ -140,7 +142,6 @@ class DefaultParser(Parser):
         self.__esc_intermediate = [] 
         self.__str = [] 
         self.__str_prefix = None 
-        self.__str_esc_state = False
         self.__timer = None
 
     def parse(self, context):
@@ -152,41 +153,64 @@ class DefaultParser(Parser):
         for c in context:
             if self.__parse_state == _STATE_OSC:
                 # parse control string
-                if c == 0x1b:
-                    self.__str_esc_state = True
-                elif c == 0x5c and self.__str_esc_state == True:
+                if c == 0x07:
                     context.dispatch_control_string(self.__str_prefix, self.__str)
                     self.__parse_state = _STATE_GROUND
-                elif c == 0x07:
-                    context.dispatch_control_string(self.__str_prefix, self.__str)
+                elif c < 0x08:
+                    seq = [0x1b] + [self.__str_prefix] + self.__str + [c]
+                    context.dispatch_invalid(seq)
                     self.__parse_state = _STATE_GROUND
+                elif c < 0x0e:
+                    self.__str.append(c)
+                elif c == 0x1b:
+                    self.__parse_state = _STATE_OSC_ESC
                 elif c < 0x20:
+                    seq = [0x1b] + [self.__str_prefix] + self.__str + [c]
+                    context.dispatch_invalid(seq)
                     self.__parse_state = _STATE_GROUND
                 else:
                     self.__str.append(c)
 
+            elif self.__parse_state == _STATE_OSC_ESC:
+                # parse control string
+                if c == 0x5c:
+                    context.dispatch_control_string(self.__str_prefix, self.__str)
+                    self.__parse_state = _STATE_GROUND
+                else:
+                    seq = [0x1b] + [self.__str_prefix] + self.__str + [0x1b, c]
+                    context.dispatch_invalid(seq)
+                    self.__parse_state = _STATE_GROUND
+
             elif self.__parse_state == _STATE_STR:
                 # parse control string
-                if c == 0x1b:
-                    self.__str_esc_state = True
-                elif c == 0x5c and self.__str_esc_state == True:
-                    context.dispatch_control_string(self.__str_prefix, self.__str)
+                # 00/08 - 00/13, 02/00 - 07/14
+                #
+                if c < 0x08:
+                    seq = [0x1b] + [self.__str_prefix] + self.__str + [c]
+                    context.dispatch_invalid(seq)
                     self.__parse_state = _STATE_GROUND
+                elif c < 0x0e:
+                    self.__str.append(c)
+                elif c == 0x1b:
+                    self.__parse_state = _STATE_STR_ESC
                 elif c < 0x20:
+                    seq = [0x1b] + [self.__str_prefix] + self.__str + [c]
+                    context.dispatch_invalid(seq)
                     self.__parse_state = _STATE_GROUND
                 else:
                     self.__str.append(c)
 
-            elif c == 0x1b: # ESC
-                self.__esc_intermediate = []
-                self.__parse_state = _STATE_ESC
-
-            elif c == 0x18 or c == 0x1a:
-                context.dispatch_char(c)
-                self.__parse_state = _STATE_GROUND
-
-            elif c < 0x20 or c == 0x7f: # control character
-                context.dispatch_char(c)
+            elif self.__parse_state == _STATE_STR_ESC:
+                # parse control string
+                # 00/08 - 00/13, 02/00 - 07/14
+                #
+                if c == 0x5c:
+                    context.dispatch_control_string(self.__str_prefix, self.__str)
+                    self.__parse_state = _STATE_GROUND
+                else:
+                    seq = [0x1b] + [self.__str_prefix] + self.__str + [0x1b, c]
+                    context.dispatch_invalid(seq)
+                    self.__parse_state = _STATE_GROUND
 
             elif self.__parse_state == _STATE_ESC:
                 #
@@ -198,12 +222,23 @@ class DefaultParser(Parser):
                 #
                 #     ESC I ... I F
                 #
-                if c == 0x5b: # [
+                if c == 0x1b: # ESC
+                    seq = [0x1b]
+                    context.dispatch_invalid(seq)
+                    self.__esc_intermediate = []
+                    self.__parse_state = _STATE_ESC
+                elif c == 0x18 or c == 0x1a:
+                    seq = [0x1b]
+                    context.dispatch_invalid(seq)
+                    context.dispatch_char(c)
+                    self.__parse_state = _STATE_GROUND
+                elif c < 0x20: # control character
+                    context.dispatch_char(c)
+                elif c == 0x5b: # [
                     self.__csi_parameter = []
                     self.__csi_intermediate = [] 
                     self.__parse_state = _STATE_CSI_PARAMETER
                 elif c == 0x5d: # ]
-                    self.__str_esc_state = False
                     self.__str = [] 
                     self.__str_prefix = c 
                     self.__parse_state = _STATE_OSC
@@ -213,7 +248,6 @@ class DefaultParser(Parser):
                     self.__parse_state = _STATE_SS3
                 elif c == 0x50 or c == 0x58 or c == 0x5e or c == 0x5f:
                     # P(DCS) or X(SOS) or ^(PM) or _(APC)
-                    self.__str_esc_state = False
                     self.__str = []
                     self.__str_prefix = c 
                     self.__parse_state = _STATE_STR
@@ -223,37 +257,70 @@ class DefaultParser(Parser):
                 elif c <= 0x7e: # ~
                     context.dispatch_esc(self.__esc_intermediate, c)
                     self.__parse_state = _STATE_GROUND
+                elif c == 0x7f: # control character
+                    context.dispatch_char(c)
                 else:
+                    seq = [0x1b]
+                    context.dispatch_invalid(seq)
                     self.__parse_state = _STATE_GROUND
-                    #raise ParseException("Unknown ESC seqnence detected.")
 
             elif self.__parse_state == _STATE_ESC_INTERMEDIATE:
-                if c <= 0x2f: # SP to /
+                if c == 0x1b: # ESC
+                    seq = [0x1b] + self.__esc_intermediate
+                    context.dispatch_invalid(seq)
+                    self.__esc_intermediate = []
+                    self.__parse_state = _STATE_ESC
+                elif c == 0x18 or c == 0x1a:
+                    seq = [0x1b] + self.__esc_intermediate
+                    context.dispatch_invalid(seq)
+                    context.dispatch_char(c)
+                    self.__parse_state = _STATE_GROUND
+                elif c < 0x20: # control character
+                    context.dispatch_char(c)
+                elif c <= 0x2f: # SP to /
                     self.__esc_intermediate.append(c)
                     self.__parse_state = _STATE_ESC_INTERMEDIATE
                 elif c <= 0x7e: # 0 to ~, Final byte
                     context.dispatch_esc(self.__esc_intermediate, c)
                     self.__parse_state = _STATE_GROUND
+                elif c == 0x7f: # control character
+                    context.dispatch_char(c)
                 else:
+                    seq = [0x1b] + self.__esc_intermediate
+                    context.dispatch_invalid(seq)
                     self.__parse_state = _STATE_GROUND
-                    #raise ParseException("Unknown ESC seqnence detected.")
 
             elif self.__parse_state == _STATE_CSI_PARAMETER:
                 # parse control sequence
                 #
                 # CSI P ... P I ... I F
                 #     ^
-                if c <= 0x2f: # intermediate, SP to /
+                if c == 0x1b: # ESC
+                    seq = [0x1b, 0x5b] + self.__csi_parameter
+                    context.dispatch_invalid(seq)
+                    self.__esc_intermediate = []
+                    self.__parse_state = _STATE_ESC
+                elif c == 0x18 or c == 0x1a:
+                    seq = [0x1b, 0x5b] + self.__csi_parameter
+                    context.dispatch_invalid(seq)
+                    context.dispatch_char(c)
+                    self.__parse_state = _STATE_GROUND
+                elif c < 0x20: # control character
+                    context.dispatch_char(c)
+                elif c <= 0x2f: # intermediate, SP to /
                     self.__csi_intermediate.append(c)
                     self.__parse_state = _STATE_CSI_INTERMEDIATE
                 elif c <= 0x3f: # parameter, 0 to ?
                     self.__csi_parameter.append(c)
                 elif c <= 0x7e: # Final byte, @ to ~
                     context.dispatch_csi(self.__csi_parameter,
-                                         self.__csi_intermediate,
-                                         c)
+                                         self.__csi_intermediate, c)
                     self.__parse_state = _STATE_GROUND
+                elif c == 0x7f: # control character
+                    context.dispatch_char(c)
                 else:
+                    seq = [0x1b, 0x5b] + self.__csi_parameter
+                    context.dispatch_invalid(seq)
                     self.__parse_state = _STATE_GROUND
                     #raise ParseException("Unknown CSI seqnence detected.")
 
@@ -262,25 +329,76 @@ class DefaultParser(Parser):
                 #
                 # CSI P ... P I ... I F
                 #             ^
-                if c <= 0x2f: # intermediate, SP to /
+                if c == 0x1b: # ESC
+                    seq = [0x1b, 0x5b] + self.__csi_parameter + self.__csi_intermediate
+                    context.dispatch_invalid(seq)
+                    self.__esc_intermediate = []
+                    self.__parse_state = _STATE_ESC
+                elif c == 0x18 or c == 0x1a:
+                    seq = [0x1b, 0x5b] + self.__csi_parameter + self.__csi_intermediate
+                    context.dispatch_invalid(seq)
+                    context.dispatch_char(c)
+                    self.__parse_state = _STATE_GROUND
+                elif c < 0x20: # control character
+                    context.dispatch_char(c)
+                elif c <= 0x2f: # intermediate, SP to /
                     self.__csi_intermediate.append(c)
                     self.__parse_state = _STATE_CSI_INTERMEDIATE
                 elif c <= 0x3f:
+                    seq = [0x1b, 0x5b] + self.__csi_parameter + self.__csi_intermediate
+                    context.dispatch_invalid(seq)
                     self.__parse_state = _STATE_GROUND
                 elif c <= 0x7e: # Final byte, @ to ~
                     context.dispatch_csi(self.__csi_parameter,
                                          self.__csi_intermediate,
                                          c)
                     self.__parse_state = _STATE_GROUND
+                elif c == 0x7f: # control character
+                    context.dispatch_char(c)
                 else:
+                    seq = [0x1b, 0x5b] + self.__csi_parameter + self.__csi_intermediate
+                    context.dispatch_invalid(seq)
                     self.__parse_state = _STATE_GROUND
             elif self.__parse_state == _STATE_SS2:
-                context.dispatch_ss2(c)
-                self.__parse_state = _STATE_GROUND
+                if c == 0x1b: # ESC
+                    seq = [0x1b, 0x4e]
+                    context.dispatch_invalid(seq)
+                    self.__esc_intermediate = []
+                    self.__parse_state = _STATE_ESC
+                elif c == 0x18 or c == 0x1a:
+                    seq = [0x1b, 0x4e]
+                    context.dispatch_invalid(seq)
+                    context.dispatch_char(c)
+                    self.__parse_state = _STATE_GROUND
+                elif c < 0x20: # control character
+                    context.dispatch_char(c)
+                elif c < 0x7f:
+                    context.dispatch_ss2(c)
+                    self.__parse_state = _STATE_GROUND
+                elif c == 0x7f: # control character
+                    context.dispatch_char(c)
             elif self.__parse_state == _STATE_SS3:
-                context.dispatch_ss3(c)
-                self.__parse_state = _STATE_GROUND
-            else:
+                if c == 0x1b: # ESC
+                    seq = [0x1b, 0x4f]
+                    context.dispatch_invalid(seq)
+                    self.__esc_intermediate = []
+                    self.__parse_state = _STATE_ESC
+                elif c == 0x18 or c == 0x1a:
+                    seq = [0x1b, 0x4f]
+                    context.dispatch_invalid(seq)
+                    context.dispatch_char(c)
+                    self.__parse_state = _STATE_GROUND
+                elif c < 0x20: # control character
+                    context.dispatch_char(c)
+                elif c < 0x7f:
+                    context.dispatch_ss3(c)
+                    self.__parse_state = _STATE_GROUND
+                elif c == 0x7f: # control character
+                    context.dispatch_char(c)
+            elif c == 0x1b: # ESC
+                self.__esc_intermediate = []
+                self.__parse_state = _STATE_ESC
+            else: # control character
                 context.dispatch_char(c)
 
         if self.__parse_state == _STATE_ESC:
@@ -383,7 +501,9 @@ class FilterMultiplexer(EventObserver):
         return handled_lhs and handled_rhs
 
     def handle_invalid(self, context, seq):
-        return False
+        handled_lhs = self.__lhs.handle_invalid(context, seq)
+        handled_rhs = self.__rhs.handle_invalid(context, seq)
+        return handled_lhs and handled_rhs
 
     def handle_draw(self, context):
         handled_lhs = self.__lhs.handle_draw(context)
@@ -524,6 +644,11 @@ class ParseContext(OutputStream, EventDispatcher):
             #else: 
             self.put(c)
 
+    def dispatch_invalid(self, seq):
+        if not self.__handler.handle_invalid(self, seq):
+            for c in seq:
+                self.put(c)
+
 ################################################################################
 #
 # DefaultPTY
